import os
import sys
import shutil
import json
import platform
import sysconfig
import argparse
import logging
import subprocess
import hashlib

from subprocess import PIPE
from collections import namedtuple

squish = None
script_version = (1, 6, 1)
python_interpreters = {}
PythonInterpreter = namedtuple('PythonInterpreter', ['version', 'architecture'])

no_revert = False
no_checks = False


def main():
    global squish

    args = parseArguments()
    setup_logging(args.verbose)
    checkArguments(args)

    squish = checkSquishLocation()
    setup_logfile(args.verbose)

    # Pydir is put into LibraryPath while pyhome replaces the value of PythonHome
    pydir, pyhome = determineInterpreterPaths()

    logging.info("Start configuring Python\n")

    pythonhomes = readSupportedPython()
    readPythonVersions(pythonhomes)

    logging.debug("Determined System => {}".format(platform.system()))
    logging.info("Used Python Directory => '{}'".format(pydir))
    logging.info("Used Python Home => '{}'\n".format(pyhome))
    logging.info("Interpreter Python Version => '{}'".format(getRunningInterpreterVersion()))
    logging.info("Interpreter Architecture => '{}'\n".format(getArchitectureString()))

    CheckRequirements()

    adjustPaths(pydir, pyhome, args.force)
    adjustLibraries(pyhome)
    checkScriptPlugin()
    logging.info("Script Execution Ended\n\n")


def parseArguments():
    parser = argparse.ArgumentParser(prefix_chars='-/',
                                     description="A script to configure an external Python in Squish")
    main_group = parser.add_argument_group()
    main_group.add_argument('--force', "-f", '/F', dest="force", action='store_true',
                            help='Forces the configuration of an interpreter if the package has support')
    revert_command_group = main_group.add_mutually_exclusive_group()
    revert_command_group.add_argument('--revert', "-R", '/R', dest="revert", action='store_true',
                                      help='Reverts to the stored backup files if any can be found')
    # Prevents Reverting to check failed configurations
    revert_command_group.add_argument('--norevert', "-nR", '/NR', dest="norevert", action='store_true', help=argparse.SUPPRESS)
    main_group.add_argument('--nochecks', "-nC", '/NC', dest="nochecks", action='store_true', help=argparse.SUPPRESS)  # Disable Linux Checks
    main_group.add_argument('--verbose', "-v", '/v', dest="verbose", action='store_true', help='Enable more logging')
    parser.add_argument('--architecture', dest="architecture", action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--version', "/V", dest="print_version", action='store_true',
                        help="Show version and hashcode of this Python Changer")
    return parser.parse_args()


def get_scripthash_value():
    hash_value = None
    try:
        with open(__file__, "rb") as rfile:
            data = rfile.read()
        hash_value = hashlib.sha256(data).hexdigest()
    except Exception as ex:
        logging.debug("Parsing Hash failed {}".format(repr(ex)))
    return hash_value


def Revertable(function):
    def inner(*args, **kwargs):
        try:
            function(*args, **kwargs)
        except RevertError as ex:
            trytoRevert(ex, ex.librevert)

    return inner


def checkArguments(args):
    execution_functions = \
        [
            (args.print_version, printVersionInfo),
            (args.revert, revertChanges),
            (args.architecture, printArchitectureInfo)
        ]

    for function_tuple in execution_functions:
        if function_tuple[0]:
            function_tuple[1]()

    config_flags = \
        {
            "no_revert": args.norevert,
            "no_checks": args.nochecks
        }

    for flag in config_flags:
        globals()[flag] = config_flags[flag]


def setup_logging(verbose):
    log_level = logging.INFO
    log_format = '%(message)s'
    if verbose:
        log_level = logging.DEBUG
    logging.basicConfig(format=log_format, level=log_level)


def setup_logfile(verbose):
    log_level = logging.INFO
    log_format = '%(levelname)s %(message)s'
    if verbose:
        log_level = logging.DEBUG
    log_formatter = logging.Formatter(log_format)
    filehandler = logging.FileHandler(os.path.join(squish, "etc", "pythchangerlog.log"))
    filehandler.setFormatter(log_formatter)
    logger = logging.getLogger()
    logger.addHandler(filehandler)
    filehandler.setLevel(log_level)


def printVersionInfo():
    logging.info("Version           : {}".format(script_version))
    logging.info("Hashcode (sha256) : {}".format(get_scripthash_value()))
    exit(0)


def printArchitectureInfo():
    print(checkArchitecture(include_pyver=True))
    exit(0)


@Revertable
def revertChanges():
    global squish
    if squish is None:
        squish = checkSquishLocation()
    raise RevertError("Argument Triggered Revert")


def checkSquishLocation():
    directory = __getSquishdir__()
    if not os.access(directory, os.W_OK):
        logging.error(
            "Could not access Squish directory '{}'\nPlease check the permissions on the folder.".format(directory))
        exit(1)
    return directory


def CheckRequirements():
    """
        For Linux check if the Python was built with shared libraries.
        Static Pythons are not supported in Squish.

        Check if a Python Virtual Environment is used and log a warning.
    """
    if no_checks:
        logging.warning("Requirement Checks will not be executed - This may lead to broken configurations\n")
        return

    virtual_environment = hasattr(sys, "real_prefix") or ( hasattr(sys, "base_prefix") and sys.prefix != sys.base_prefix)
    logging.debug("Virtual Environment Check - {} / base_prefix:{} / real_prefix:{} ".format(virtual_environment,
                                                                                             hasattr(sys, "base_prefix"),
                                                                                             hasattr(sys, "real_prefix")))
    if virtual_environment:
        logging.warning("WARNING :\nDetected Virtual Environment in Configuration\n"
                        "This configuration may work or be unusable due to how Python virtual environments work.\n"
                        "We recommend the use of a standard Python\n")
        logging.debug("Environment : sys.prefix-> {} - sys.base_prefix-> {}".format(sys.prefix, sys.base_prefix))

    if platform.system() == 'Linux':
        configvars = sysconfig.get_config_var('CONFIG_ARGS')
        if configvars is not None:
            if '--enable-shared' not in configvars:
                logging.error("Python has been statically built and will not work for Squish")
                logging.debug("CONFIG_ARGS -> {}".format(configvars))
                sys.exit(1)

        shared_lib = "libpython{}.so".format(sys.version_info.major)
        lib_info = sysconfig.get_config_vars('LIBDIR', 'INSTSONAME')
        logging.debug("Libinfo {}".format(lib_info))
        libname = lib_info[1]
        logging.debug("Looking for shared library '{}' or API '{}'".format(libname, shared_lib))
        cmd = "ldconfig -p | grep -e {} -e {}".format(libname, shared_lib)
        try:
            output = subprocess.check_output(cmd, shell=True)
        except subprocess.CalledProcessError:
            output = ""
        output = str(output).strip("b'\\t").rstrip('\\n')
        split = output.split("\\n\\t")
        if len(split) > 1:
            output = '\n'.join(split)
        logging.debug("Output from '{}'->\n{}\n".format(cmd, output))
        lib_path = os.sep.join(lib_info)
        logging.debug("Shared Library {} -> Exists {}".format(lib_path, os.path.isfile(lib_path)))
        if output == "" and not os.path.isfile(lib_path):
            logging.error("Could not find shared library '{}'\n"
                          "The Python library is required for Squish to work with this Python installation\n"
                          .format(libname))
            logging.info('Script execution aborted')
            exit(-1)


def adjustLibraries(pyhome):
    """
    Only required for Mac and Windows(if Python 2 Interpreter of Squish is older than the one configured)
    """
    if platform.system() == "Darwin":
        changeMacLibraryPaths(pyhome)
    elif platform.system() == "Windows":
        hidePython2Library()


def __getSquishdir__():
    for direc in [os.getcwd(), os.path.dirname(os.path.abspath(__file__))]:
        buildinfo_path = os.path.join(direc, "buildinfo.txt")
        if os.path.isfile(buildinfo_path) \
                and os.path.getsize(buildinfo_path) > 0 \
                and os.path.isfile(os.path.join(direc, "etc", "paths.ini")):
            return direc
    else:
        logging.error("Unable to find valid marker files buildinfo.txt and etc/paths.ini.\n"
                      "Please execute the script from the installation directory of a Squish binary package\n")
        logging.info("Script execution aborted")
        exit(1)


@Revertable
def checkScriptPlugin():
    logging.info("Checking if Python Script Plugin can be loaded")
    if platform.system() == "Windows":
        envvalues = {"SystemRoot": os.environ["SystemRoot"], "USERPROFILE": os.environ["USERPROFILE"]}
    else:
        envvalues = {"HOME": os.environ["HOME"]}
    output = subprocess.Popen([os.path.join(squish, "bin", "squishrunner"), "--scriptLanguages"], stderr=PIPE,
                              stdout=PIPE, env=envvalues).communicate()
    logging.debug("Output from --scriptLanguages:\n{}".format(output[0].decode('utf-8')))
    if "Python" in str(output[0]):
        logging.info("Plugin loaded successful")
        logging.info("Configuration Finished")
    else:
        logging.warning("Plugin loading failed - Trying to revert changes")
        raise RevertError("Loading of script plugin failed")


def readSupportedPython():
    """
    Read Python Homes from the setup.ini file
    """
    homes = []
    with open(os.path.join(squish, "etc", "setup.ini")) as rfile:
        lines = rfile.readlines()

    for line in lines:
        if line[0] == '[':
            if "Python" not in line:
                continue

        line = line.rstrip(os.linesep)
        splitline = [x.strip(' ').strip('"') for x in line.split('=')]
        if splitline[0] == "Home":
            homes.append(splitline[1].replace("@(SQUISH_PREFIX)", squish).replace("\\\\", "\\"))
    return homes


def readPythonVersions(pythonhomes):
    """
    Poll Python Interpreters from the Squish package for their architecture and version.
    """
    dirlist = [direc for direc in os.listdir(squish) if direc.startswith('python')]
    if len(dirlist) == 0:
        logging.error("No Python directories found."
                      "This could mean that either the Python interpreter has been removed"
                      "or that this is a source build.\n"
                      "This script is only intended for Squish binary packages provided by the vendor.")
        exit(0)

    if len(pythonhomes) == 0:
        logging.info(r"No pythonhomes found in <SQUISHDIR>\etc\setup.ini "
                     "- Older package falling back to using directories in <SQUISHDIR>")
        pythonhomes += dirlist

    global python_interpreters
    for directory in pythonhomes:
        try:
            if platform.system() == "Windows":
                pybinary = os.path.join(squish, directory, "python.exe")
            else:
                pybinary = os.path.join(squish, directory, "bin", "python")

            if not os.path.isfile(pybinary):
                logging.warning("The Python binary '{}' does not exist\n"
                                "Either the binary was deleted or this is a source build\n"
                                .format(pybinary.replace(squish, "<SQUISHDIR>")))
                continue

            output = subprocess.check_output([pybinary, __file__, '--architecture'])
            output = json.loads(output.decode("utf-8").replace("'", '"'))
            interpreter_name = output[0]
            arch = ' '.join(output[1:])
            python_interpreters[directory.replace(squish, "<SQUISHDIR>")] = PythonInterpreter(interpreter_name, arch)
        except Exception as ex:
            logging.warning("Exception during Python checking of directory '{}'\n{}".format(directory, ex))
            logging.debug("Output from command : '{}'".format(output))
            python_interpreters[directory.replace(squish, "<SQUISHDIR>")] = PythonInterpreter("NaN", "NaN")


def getArchitectureString(include_pyver=False):
    return ' '.join(checkArchitecture(include_pyver))


def checkArchitecture(include_pyver=False):
    architecture = []
    if include_pyver:
        architecture.append('{}.{}'.format(sys.version_info.major, sys.version_info.minor))
    if platform.system() == "Darwin":
        architecture.append(platform.machine())
    architecture.append(__check_bitness())
    return architecture


def __check_bitness():
    if sys.maxsize > 2 ** 32:
        return "64-bit"
    else:
        return "32-bit"


def determineInterpreterPaths():
    if platform.system() == "Windows":
        bindir = os.path.dirname(sys.executable)
        yield bindir
        yield bindir
    elif platform.system() == "Linux":
        yield os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.__file__))), "lib")
        yield os.path.dirname(os.path.dirname(sys.executable))
    elif platform.system() == "Darwin":  # macOS
        yield os.path.dirname(sys.executable)
        yield sys.prefix
    else:
        logging.error("Unknown system '{}'".format(platform.system()))
        exit(1)


def adjustPaths(pydir, pyhome, force_config):
    pathsfile = os.path.join(squish, "etc", "paths.ini")
    with open(pathsfile, "r") as rfile:
        lines = rfile.readlines()

        try:
            Python_Combi = lines[1].split('"')[1]
            logging.info("Active Python in Package - '{}'".format(Python_Combi))
        except IndexError:
            Python_Combi = getRunningInterpreterVersion()
    logging.info("Found Python Versions =>\n'{}'\n\n".format(python_interpreters))

    if checkPythonVersion(python_interpreters, Python_Combi):
        logging.info("Found matching Python in package for Python interpreter")
        updatePathsFile(pathsfile, lines, pydir, pyhome)
    elif checkPythonVersion(python_interpreters, getRunningInterpreterVersion()):
        logging.info("Package contains support for the Python version but is configured for a different Python version.")
        if force_config:
            logging.info("Enforcing Configuration")
            updatePathsFile(pathsfile, lines, pydir, pyhome, True)
            return
        logging.info("If you wish to update the configuration anyway,"
                     " please re-run the script with --force or /F argument")
        exit(1)
    else:
        logging.info("The Squish package does not support the interpreter this script was started with.\n")
        checkFailureReason()
        exit(0)


def checkFailureReason():
    package_architecture = [p for p in python_interpreters.values()][0].architecture
    interpreter_architecture = getArchitectureString()
    if package_architecture == interpreter_architecture:

        interpreter_version = getRunningInterpreterVersion()
        supported_versions = [p.version[:-(len(p.version) - p.version.rindex('.'))] for p in
                              python_interpreters.values()]
        if any([version in interpreter_version for version in supported_versions]):
            logging.info("The Squish package has no support for Python {0}"
                         " and cannot be configured for this interpreter.\n"
                         "The package contains support for the following Python version(s): "
                         " {1}\nPlease login to your Qt Account at https://login.qt.io/login to inquire about support for Python {0}"
                         " if you want to use that version."
                         .format(interpreter_version, " & ".join(["{}.x".format(x.version) for x in python_interpreters.values()])))
        else:
            logging.info("Unknown Error")
            logging.debug("Architecture Check => {}".format(package_architecture == interpreter_architecture))
            logging.debug("Interpreter Version => {}".format(interpreter_version))
            logging.debug("Supported Versions => {}".format(" - ".join(supported_versions)))
    else:
        logging.info("The Squish package only has support for {} Python interpreters.\n"
                     "If you require a package for {} Python"
                     " please contact our support by logging in to your Qt Account at https://login.qt.io/login to request such a package."
                     .format(package_architecture, interpreter_architecture))


def hidePython2Library():
    """
        Python 2  - Only on Windows
        Due to older python27.dll libraries breaking tests, a python27.dll in Path was necessary.
        Python 2 is on the way out, just hide the library always.
    """
    if sys.version_info.major == 2:
        source = os.path.join(squish, "lib", "python27.dll")
        dest = os.path.join(squish, "lib", "python27_off.dll")

        if not os.path.exists(source):
            logging.warning("[Warning] - Could not find python27.dll at '{}'".format(source))
            return

        try:
            shutil.move(source, dest)
            logging.debug("Moved library file\n{} => {}".format(source,dest))
        except IOError as ex:
            raise ex


def restorePython2Library():
    if sys.version_info.major == 2:
        source = os.path.join(squish, "lib", "python27_off.dll")
        dest = os.path.join(squish, "lib", "python27.dll")
        try:
            shutil.move(source, dest)
            logging.debug("Restored Python 2 library\n{} => {}".format(source,dest))
        except IOError as ex:
            logging.debug("IOException {}".format(ex))
            if 'No such' not in str(ex):
                raise ex


@Revertable
def changeMacLibraryPaths(pyhome):
    squishpypath = os.path.join(squish, "lib",
                                "libsquishpython-{}.{}.dylib".format(sys.version_info.major, sys.version_info.minor))
    runInstallNameTool(pyhome, squishpypath)

    squishtestpath = os.path.join(squish, "lib", "squishtest{}.so".format(sys.version_info.major))
    runInstallNameTool(pyhome, squishtestpath)


def runInstallNameTool(pyhome, filepath):
    """
    For Mac the Pathnames in the libraries need to be patched. This requires XCode to be present on the system
    """

    def _invoke_install_name_tool(src_install_name, dst_install_name, filepath):
        cmd = 'install_name_tool -change {} "{}" "{}"'.format(src_install_name, dst_install_name, filepath)
        code = os.system(cmd)
        outputpath = os.path.join("<SQUISHDIR>", filepath.split(os.sep)[-2], os.path.basename(filepath))
        if code == 0:
            logging.debug("Finished updating library names from {} to {} for {}".format(src_install_name,
                                                                                        dst_install_name,
                                                                                        outputpath))
        else:
            logging.error("Failed to write library names")
            logging.info("Errorcode => '{}'".format(code))
            logging.info("Command => '{}'".format(cmd))
            raise RevertError("Reverting due to failing to update library names")

    if os.system("xcode-select -p > null") == 0:
        libname = "libpython{}.{}.dylib".format(sys.version_info.major, sys.version_info.minor)
        pythonname = "@rpath/{}".format(libname)
        pybinary = "{}/Python".format(pyhome)
        if (not os.path.exists(pybinary)) and sys.version_info.major == 3:
            pybinary = "{}/Python3".format(pyhome)
        _invoke_install_name_tool(pythonname, pybinary, filepath)
        _invoke_install_name_tool("@loader_path/../python3/Python.framework/Versions/Current/Python", pybinary,
                                  filepath)
    else:
        logging.error("Xcode is missing from this macOS system. "
                      "Please install it as it is required for some procedures.")
        raise RevertError("Xcode is not installed")


def getRunningInterpreterVersion():
    return "{}.{}".format(sys.version_info.major, sys.version_info.minor)


def updatePathsFile(source, lines, pydir, pyhome, updateVersion=False):
    """
    Update the paths.ini file for the new Settings.
    """
    output = []
    if platform.system() == "Windows":
        sep = "\\\\"
        pydir = pydir.replace("\\", sep)
    else:
        sep = "/"

    repkeys = ["@(SQUISH_PREFIX){}python".format(sep), "/Library/Frameworks/Python.framework"]

    for line in lines:
        splitline = line.split("=")
        if len(splitline) > 1:
            key = splitline[0].rstrip(" ")

            if key in ["PythonPluginVersion", "LibraryPath", "Scripting/PythonHome"]:
                if key == "PythonPluginVersion" and updateVersion:
                    line = 'PythonPluginVersion = "{}"{}'.format(getRunningInterpreterVersion(), "\n")

                if key == "LibraryPath":
                    value = splitline[1].lstrip(" ").strip("\"\n")

                    if not any([rkey in value for rkey in repkeys]):
                        ExitModifiedPaths()

                    split = value.split(os.pathsep)
                    for item in split:
                        if any([rkey in item for rkey in repkeys]):
                                split[split.index(item)] = pydir
                    split = list(set(split))
                    value = os.pathsep.join(split)
                    line = '{} = "{}"\n'.format(key, value)

                if key == "Scripting/PythonHome":
                    value = splitline[1].lstrip(" ")

                    if not any([rkey in value for rkey in repkeys]):
                        ExitModifiedPaths()

                    value = '"{}"'.format(pyhome.replace(os.sep, sep))
                    line = "{} = {}{}".format(key, value, "\n")

        output.append(line)

    performBackups()

    try:
        directory = os.path.dirname(source)
        filename = os.path.basename(source)
        filepath = os.path.join(directory, "{}.temp".format(filename))
        WriteTempfile(filepath, output)
        logging.debug("Tempfile was created successfully")

        shutil.move(filepath, source)
        logging.debug("Moved Tempfile\n{} -> {}".format(filepath, source))
    except IOError as ex:
        logging.error("Moving Temp File Failed\n{}".format(ex.args))
        exit(1)
    logging.info("Finished Writing Configuration\n")


def WriteTempfile(filepath, lines):
    try:
        with open(filepath, 'w') as wfile:
            for line in lines:
                wfile.write(line)
    except IOError as ex:
        logging.error('Could not create tempfile {}\nReason -> {}'.format(filepath, ex))


def ExitModifiedPaths():
    logging.error("The file {} was modified before. Please revert the changes and run the script again.".format(
        os.path.join("<SQUISHDIR>", "etc", "paths.ini")))
    exit(1)


def checkPythonVersion(versions, Python_Combi):
    if getRunningInterpreterVersion() in Python_Combi:
        architecture = getArchitectureString()
        for ver in versions:
            logging.debug('Python Version {}'.format(ver))
            if getRunningInterpreterVersion() in versions[ver].version and architecture == versions[ver].architecture:
                return ver
    return None


def trytoRevert(exc, LibRevert=True):
    """
    Revert to a previous Setup if a backup exists
    """

    if no_revert:
        logging.warning("Revert aborted by argument")
        return

    logging.info("Reverting to Backup - Reason '{}'".format(exc.message))

    source = os.path.join(squish, "etc", "paths.ini")
    revertBackup(source)

    if platform.system() == 'Darwin' and LibRevert:
        revertLibraries()
    if platform.system() == 'Windows':
        restorePython2Library()

    logging.info("Changes reverted")
    exit(0)


def revertLibraries():
    original = os.path.join(squish, "lib",
                            "libsquishpython-{}.{}.dylib".format(sys.version_info.major, sys.version_info.minor))
    revertBackup(original)

    squishtestpath = os.path.join(squish, "lib", "squishtest{}.so".format(sys.version_info.major))
    revertBackup(squishtestpath)


@Revertable
def performBackups():
    pathsfile = os.path.join(squish, "etc", "paths.ini")
    copyBackup(pathsfile)

    if platform.system() == "Darwin":
        try:
            backupLibraries()
        except Exception as ex:
            raise RevertError(ex.args, False)


def backupLibraries():
    original = os.path.join(squish, "lib",
                            "libsquishpython-{}.{}.dylib".format(sys.version_info.major, sys.version_info.minor))
    copyBackup(original, errormode='strict')

    squishtestpath = os.path.join(squish, "lib", "squishtest{}.so".format(sys.version_info.major))
    copyBackup(squishtestpath, errormode='strict')


def copyBackup(filepath, errormode='relaxed'):
    backup = "{}.original".format(filepath)
    outputpath = os.path.join("<SQUISHDIR>", filepath.split(os.sep)[-2], os.path.basename(filepath))
    try:
        shutil.copy(filepath, backup)
        logging.info("Performed Backup for {}".format(outputpath))
    except IOError as ex:
        logging.error("Could not copy backup for {}\nException {}".format(outputpath,ex))
        exception = ex

    try:
        exception
    except NameError:
        return
    else:
        if errormode == 'strict':
            raise exception
    exit(1)


def revertBackup(filepath):
    backup = "{}.original".format(filepath)
    outputpath = os.path.join("<SQUISHDIR>", filepath.split(os.sep)[-2], os.path.basename(filepath))
    try:
        shutil.move(backup, filepath)
        logging.info("Reverted to backup file for {}".format(outputpath))
    except IOError as ex:
        logging.error("Could not revert to backup for {}\nException {}".format(outputpath,ex))
        exit(1)

class RevertError(Exception):
    librevert = True

    def __init__(self, message, librevert=True):
        self.message = message
        self.librevert = librevert


if __name__ == '__main__':
    main()
